/*
 * kgraft_patch_bsc1071471_cmtp
 *
 * Fix for CVE-2017-15868, bsc#1071471 -- cmtp part
 *
 *  Upstream commit:
 *  96c26653ce65 ("Bluetooth: cmtp: cmtp_add_connection() should verify that
 *                 it's dealing with l2cap socket")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP2 commit:
 *  not affected
 *
 *  SLE12(-SP1) commit:
 *  cdf18858dc3600221c3ad6f1c98f55d0e288512a
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/socket.h>
#include <linux/skbuff.h>
#include <net/bluetooth/bluetooth.h>
#include <net/bluetooth/l2cap.h>
#include <linux/kthread.h>
#include <linux/isdn/capilli.h>

#if !IS_MODULE(CONFIG_BT_CMTP)
#error "KGR patch supports only CONFIG_BT_CMTP=m."
#endif

#if !IS_MODULE(CONFIG_BT)
#error "KGR patch supports only CONFIG_BT=m."
#endif

#define KGR_PATCHED_MODULE "cmtp"

static struct module *kgr_cmtp_module;

static struct list_head *kgr_cmtp_session_list;
static struct rw_semaphore *kgr_cmtp_session_sem;

static bool (*kgr_l2cap_is_socket)(struct socket *sock);
static struct cmtp_session* (*kgr__cmtp_get_session)(bdaddr_t *bdaddr);
static int (*kgr_cmtp_session)(void *arg);
static int (*kgr_cmtp_attach_device)(struct cmtp_session *session);

static struct {
       char *name;
       void **addr;
} kgr_funcs[] = {
	{ "cmtp:cmtp_session_list", (void *)&kgr_cmtp_session_list },
	{ "cmtp:cmtp_session_sem", (void *)&kgr_cmtp_session_sem },
	{ "bluetooth:l2cap_is_socket", (void *)&kgr_l2cap_is_socket },
	{ "cmtp:__cmtp_get_session", (void *)&kgr__cmtp_get_session },
	{ "cmtp:cmtp_session", (void *)&kgr_cmtp_session },
	{ "cmtp:cmtp_attach_device", (void *)&kgr_cmtp_attach_device },
};

/* from net/bluetooth/cmtp/cmtp.h */
#define BTNAMSIZ 18
#define CMTP_LOOPBACK	0

struct cmtp_connadd_req {
	int   sock;	/* Connected socket */
	__u32 flags;
};

#define CMTP_INITIAL_MSGNUM	0xff00

struct cmtp_session {
	struct list_head list;

	struct socket *sock;

	bdaddr_t bdaddr;

	unsigned long state;
	unsigned long flags;

	uint mtu;

	char name[BTNAMSIZ];

	atomic_t terminate;
	struct task_struct *task;

	wait_queue_head_t wait;

	int ncontroller;
	int num;
	struct capi_ctr ctrl;

	struct list_head applications;

	unsigned long blockids;
	int msgnum;

	struct sk_buff_head transmit;

	struct sk_buff *reassembly[16];
};


/* from net/bluetooth/cmtp/core.c */
/* inlined */
static void kgr__cmtp_link_session(struct cmtp_session *session)
{
	list_add(&session->list, kgr_cmtp_session_list);
}

/* inlined */
static void kgr__cmtp_unlink_session(struct cmtp_session *session)
{
	list_del(&session->list);
}


/* patched */
int kgr_cmtp_add_connection(struct cmtp_connadd_req *req, struct socket *sock)
{
	struct cmtp_session *session, *s;
	int i, err;

	BT_DBG("");

	/*
	 * Fix CVE-2017-15868
	 *  +3 lines
	 */
	if (!kgr_l2cap_is_socket(sock))
		return -EBADFD;

	session = kzalloc(sizeof(struct cmtp_session), GFP_KERNEL);
	if (!session)
		return -ENOMEM;

	down_write(kgr_cmtp_session_sem);

	s = kgr__cmtp_get_session(&bt_sk(sock->sk)->dst);
	if (s && s->state == BT_CONNECTED) {
		err = -EEXIST;
		goto failed;
	}

	bacpy(&session->bdaddr, &bt_sk(sock->sk)->dst);

	session->mtu = min_t(uint, l2cap_pi(sock->sk)->chan->omtu,
					l2cap_pi(sock->sk)->chan->imtu);

	BT_DBG("mtu %d", session->mtu);

	sprintf(session->name, "%pMR", &bt_sk(sock->sk)->dst);

	session->sock  = sock;
	session->state = BT_CONFIG;

	init_waitqueue_head(&session->wait);

	session->msgnum = CMTP_INITIAL_MSGNUM;

	INIT_LIST_HEAD(&session->applications);

	skb_queue_head_init(&session->transmit);

	for (i = 0; i < 16; i++)
		session->reassembly[i] = NULL;

	session->flags = req->flags;

	kgr__cmtp_link_session(session);

	__module_get(kgr_cmtp_module);
	session->task = kthread_run(kgr_cmtp_session, session, "kcmtpd_ctr_%d",
								session->num);
	if (IS_ERR(session->task)) {
		module_put(kgr_cmtp_module);
		err = PTR_ERR(session->task);
		goto unlink;
	}

	if (!(session->flags & (1 << CMTP_LOOPBACK))) {
		err = kgr_cmtp_attach_device(session);
		if (err < 0) {
			atomic_inc(&session->terminate);
			wake_up_process(session->task);
			up_write(kgr_cmtp_session_sem);
			return err;
		}
	}

	up_write(kgr_cmtp_session_sem);
	return 0;

unlink:
	kgr__cmtp_unlink_session(session);

failed:
	up_write(kgr_cmtp_session_sem);
	kfree(session);
	return err;
}


static int kgr_patch_bsc1071471_cmtp_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1071471_cmtp_module_notify(struct notifier_block *nb,
					unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	kgr_cmtp_module = mod;

	ret = kgr_patch_bsc1071471_cmtp_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1071471_cmtp_module_nb = {
	.notifier_call = kgr_patch_bsc1071471_cmtp_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1071471_cmtp_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	kgr_cmtp_module = find_module(KGR_PATCHED_MODULE);
	if (kgr_cmtp_module) {
		ret = kgr_patch_bsc1071471_cmtp_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1071471_cmtp_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1071471_cmtp_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1071471_cmtp_module_nb);
}
